import 'dotenv/config';
import chalk from 'chalk';
import { Octokit, App } from 'octokit';
import fetch from 'node-fetch';
import open from 'open';
import crypto from 'crypto';
import * as fs from 'node:fs';
import { LAUNCHPAD_API_ROOT, LAUNCHPAD_BUG_STATUSES, GITHUB_LABELS, LAUNCHPAD_CONSUMER_KEY } from '#lib/constants.js';

const migrationStartTime = new Date();

const sleep = (ms) => new Promise((r) => setTimeout(r, ms));

function waitForKey(keyCode) {
    return new Promise((resolve, reject) => {
        process.stdin.on('data', function (chunk) {
            if (chunk[0] === keyCode) {
                resolve();
            }
        });
    });
}

let configData = JSON.stringify({ firstRun: true });
if (process.env?.FORCE_EXECUTION == 'true') {
    fs.writeFileSync('config.json', configData);
}

const rawdata = fs.readFileSync('config.json');
const config = JSON.parse(rawdata);

configData = JSON.stringify({ firstRun: false });
if (config?.firstRun === true) {
    fs.writeFileSync('config.json', configData);
} else {
    console.log(chalk.black.bgRed(' This script has already been executed '));
    process.exit(1);
}

console.log(chalk.black.bgYellow(' Starting Launchpad to GitHub migration... '), '\n');
process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Verifying GitHub and Launchpad data\r`);

// STEP 0: Validate the GitHub access token, repository and the Launchpad project
const octokit = new Octokit({
    auth: process.env.GITHUB_ACCESS_TOKEN,
});

try {
    // Check whether the provided repository exists
    const repository = await octokit.request('GET /repos/{owner}/{repo}', {
        owner: process.env.REPOSITORY_OWNER,
        repo: process.env.REPOSITORY_NAME,
    });
} catch (e) {
    const errors = { 404: `The GitHub repository "${process.env.REPOSITORY_OWNER}/${process.env.REPOSITORY_NAME}" was not found`, 401: 'Invalid GitHub authentication token' };
    process.stdout.clearLine();
    console.log(`${chalk.black.bgRed(' FAILED ')} ${errors?.[e.status]}`);
    process.exit(1);
}

let launchpadProject = await fetch(LAUNCHPAD_API_ROOT + process.env.LAUNCHPAD_PROJECT_NAME);
if (launchpadProject.status !== 200) {
    process.stdout.clearLine();
    console.log(`${chalk.black.bgRed(' FAILED ')} Invalid Launchpad project`);
    process.exit(1);
} else {
    launchpadProject = await launchpadProject.json();
    process.stdout.clearLine();
    console.log(`${chalk.black.bgGreen(' SUCCESS ')} Verifying GitHub and Launchpad data`);
}

// STEP 1: Authenticate with Launchpad
process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Preparing Launchpad authentication\r`);
const requestTokenResponse = await (
    await fetch(`https://launchpad.net/+request-token`, {
        method: 'POST',
        headers: {
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
        },
        body: new URLSearchParams({
            oauth_consumer_key: LAUNCHPAD_CONSUMER_KEY,
            oauth_signature_method: 'PLAINTEXT',
            oauth_signature: '&',
        }),
    })
).text();
const requestToken = new URLSearchParams(requestTokenResponse);

process.stdout.write(`${chalk.black.bgYellow(' ACTION REQUIRED ')} To continue, you must authenticate with Launchpad. Press the ENTER key to be taken to Launchpad.\r`);
// Wait for ENTER keypress
await waitForKey(13);
open(`https://launchpad.net/+authorize-token?oauth_token=${requestToken.get('oauth_token')}`);
process.stdout.write(`${chalk.black.bgYellow(' ACTION REQUIRED ')} Once you've authorized this script on Launchpad, press the ENTER key.\r`);
// Wait again for ENTER keypress
await waitForKey(13);

const accessTokenResponse = await (
    await fetch(`https://launchpad.net/+access-token`, {
        method: 'POST',
        headers: {
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
        },
        body: new URLSearchParams({
            oauth_token: requestToken.get('oauth_token'),
            oauth_consumer_key: LAUNCHPAD_CONSUMER_KEY,
            oauth_signature_method: 'PLAINTEXT',
            oauth_signature: '&' + requestToken.get('oauth_token_secret'),
        }),
    })
).text();
const accessToken = new URLSearchParams(accessTokenResponse);

if (accessToken.get('oauth_token')) {
    console.log(`${chalk.black.bgGreen(' SUCCESS ')} Successful Launchpad authorization`);
} else {
    console.log(`${chalk.black.bgRed(' FAILED ')} Failed Launchpad authorization`);
    process.exit(1);
}

// STEP 2: Create the necessary GitHub labels
async function createGitHubLabels(labels) {
    for (const label in labels) {
        try {
            await octokit.request('POST /repos/{owner}/{repo}/labels', {
                owner: process.env.REPOSITORY_OWNER,
                repo: process.env.REPOSITORY_NAME,
                name: labels[label].name,
                color: labels[label].color,
            });
        } catch {}
        await sleep(500);
    }
}

process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Creating GitHub labels\r`);
await createGitHubLabels(GITHUB_LABELS.miscellaneous);
await createGitHubLabels(GITHUB_LABELS.priority);
await createGitHubLabels(GITHUB_LABELS.status);
//await createGitHubLabels(GITHUB_LABELS.definition);
//await createGitHubLabels(GITHUB_LABELS.implementation);
process.stdout.clearLine();
console.log(`${chalk.black.bgGreen(' SUCCESS ')} Creating GitHub labels`);

// STEP 3: Transfer milestones from Launchpad to GitHub
async function createGitHubMilestone(milestone) {
    try {
        await octokit.request('POST /repos/{owner}/{repo}/milestones', {
            owner: process.env.REPOSITORY_OWNER,
            repo: process.env.REPOSITORY_NAME,
            title: milestone.name,
            state: milestone.is_active === true ? 'open' : 'closed',
            description: milestone.title,
            ...(milestone.date_targeted && { due_on: new Date(milestone.date_targeted).toISOString() }),
        });
    } catch {}
    await sleep(500);
}

process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Creating GitHub milestones\r`);
const launchpadMilestones = await (await fetch(launchpadProject.all_milestones_collection_link)).json();
for (const milestone of launchpadMilestones.entries) {
    await createGitHubMilestone(milestone);
}

let milestones = {};
const githubMilestones = await octokit.request('GET /repos/{owner}/{repo}/milestones{?state,sort,direction,per_page,page}', {
    owner: process.env.REPOSITORY_OWNER,
    repo: process.env.REPOSITORY_NAME,
    state: 'all',
    per_page: 100,
});
// Find the GitHub milestone number of each Launchpad milestone
for (const milestone of githubMilestones.data) {
    const milestoneData = launchpadMilestones.entries.find((_milestone) => _milestone.name === milestone.title);
    milestones[milestoneData.name] = milestone.number;
}

process.stdout.clearLine();
console.log(`${chalk.black.bgGreen(' SUCCESS ')} Creating GitHub milestones`);

// STEP 4: Retrieve all Launchpad bugs
let bugs = [];

async function getLaunchpadBugs() {
    const bugStatuses = [...LAUNCHPAD_BUG_STATUSES.open, ...LAUNCHPAD_BUG_STATUSES.closed.completed, ...LAUNCHPAD_BUG_STATUSES.closed.notPlanned];
    const bugStatusesURIFormat = `["${bugStatuses.join('","')}"]`;

    const totalBugs = Number(await (await fetch(`${LAUNCHPAD_API_ROOT}${process.env.LAUNCHPAD_PROJECT_NAME}?status=${bugStatusesURIFormat}&ws.op=searchTasks&ws.size=100&ws.show=total_size`)).text());

    let offset = 0; // **IMPORTANT** Change this if you need to skip over already migrated bugs

    async function getBugs(offset) {
        const res = await (await fetch(`${LAUNCHPAD_API_ROOT}${process.env.LAUNCHPAD_PROJECT_NAME}?status=${bugStatusesURIFormat}&ws.op=searchTasks&ws.size=100&ws.start=${offset}`)).json();
        bugs = [...bugs, ...res.entries];
        offset += 100;

        if (res.entries.length < 100) return;
        await getBugs(offset);
    }

    await getBugs(offset);
}

process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Retrieving Launchpad bugs\r`);
await getLaunchpadBugs();
process.stdout.clearLine();
console.log(`${chalk.black.bgGreen(' SUCCESS ')} Retrieving Launchpad bugs`);

// STEP 5: Get Launchpad bug details, including comments
const statuses = {
    Invalid: GITHUB_LABELS.status.invalid,
    "Won't Fix": GITHUB_LABELS.status.wontFix,
    Expired: GITHUB_LABELS.status.expired,
    Opinion: GITHUB_LABELS.status.opinion,
    Confirmed: GITHUB_LABELS.status.confirmed,
    Triaged: GITHUB_LABELS.status.triaged,
    'In Progress': GITHUB_LABELS.status.inProgress,
    'Fix Committed': GITHUB_LABELS.status.fixCommitted,
    'Fix Released': GITHUB_LABELS.status.fixReleased,
};

const priorities = {
    Undecided: GITHUB_LABELS.priority.undecided,
    Undefined: GITHUB_LABELS.priority.undecided,
    Wishlist: GITHUB_LABELS.priority.wishlist,
    Not: GITHUB_LABELS.priority.wishlist,
    Low: GITHUB_LABELS.priority.low,
    Medium: GITHUB_LABELS.priority.medium,
    High: GITHUB_LABELS.priority.high,
    Critical: GITHUB_LABELS.priority.critical,
    Essential: GITHUB_LABELS.priority.critical,
};

async function getLaunchpadBugComments(launchpadComments) {
    let comments = [];

    for (const comment of launchpadComments) {
        const commentOwner = await (await fetch(comment.owner_link)).json();

        comments.push({
            created_at: new Date(comment.date_created).toISOString(),
            body: `
#### Imported from ${comment.web_link}
|Property|Value|
|----|----|
|Posted by|[${commentOwner?.display_name} (${commentOwner.name})](${commentOwner.web_link})|
|Date posted|${new Date(comment.date_created).toUTCString()}|

${comment.content}`,
        });
    }

    return comments;
}

let githubIssuesToImport = [];
let launchpadBugIds = [];

async function getLaunchpadBugDetails(bug) {
    const bugData = await (await fetch(bug.bug_link)).json();
    const bugOwner = await (await fetch(bug.owner_link)).json();
    const bugTask = await (await fetch(bugData.bug_tasks_collection_link)).json();
    const bugMilestone = bugTask?.entries?.[0]?.milestone_link ? await (await fetch(bugTask?.entries?.[0]?.milestone_link)).json() : undefined;
    const bugComments = bugData.message_count > 1 ? await (await fetch(bugData.messages_collection_link)).json() : undefined;

    const bugStatus = bugTask?.entries?.[0]?.status;
    const bugPriority = bugTask?.entries?.[0]?.importance;

    let comments = [];

    if (bugComments) {
        comments = await getLaunchpadBugComments(bugComments.entries.slice(1));
    }

    githubIssuesToImport.push({
        issue: {
            title: `[BUG n°${bugData.id}] ${bugData.title}`,
            body: `
#### Imported from ${bugData.web_link}
|Property|Value|
|----|----|
|Reported by|[${bugOwner?.display_name} (${bugOwner.name})](${bugOwner.web_link})|
|Date reported|${new Date(bugData.date_created).toUTCString()}|
${bugData.tags.length > 0 ? `|Tags|${bugData.tags.join(', ')}|` : ''}

${bugData.description}`,
            created_at: new Date(bugData.date_created).toISOString(),
            updated_at: new Date(bugData.date_last_updated).toISOString(),
            ...(bugMilestone && { milestone: milestones[bugMilestone.name] }),
            closed: [...LAUNCHPAD_BUG_STATUSES.closed.completed, ...LAUNCHPAD_BUG_STATUSES.closed.notPlanned].includes(bugStatus),
            ...(bugTask?.entries?.[0]?.date_closed && { closed_at: new Date(bugTask.entries[0].date_closed).toISOString() }),
            labels: [
                GITHUB_LABELS.miscellaneous.bug.name,
                GITHUB_LABELS.miscellaneous.launchpadTransfer.name,
                ...(bugStatus && statuses?.[bugStatus] ? [statuses[bugStatus].name] : []),
                ...(bugPriority && priorities?.[bugPriority] ? [priorities[bugPriority].name] : []),
            ],
        },
        comments: comments,
    });

    launchpadBugIds.push(bugData.id);
}

for (const [i, bug] of bugs.entries()) {
    process.stdout.clearLine();
    process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Retrieving Launchpad bug details. Completed: ${i}/${bugs.length}\r`);
    await getLaunchpadBugDetails(bug);
}
process.stdout.clearLine();
console.log(`${chalk.black.bgGreen(' SUCCESS ')} Retrieved details and comments for ${bugs.length} Launchpad bugs`);

// STEP 6: Transfer Launchpad bugs, one by one, to GitHub issues
let pendingIssues = [];
let launchpadBugsThatCouldNotBeImported = [];

for (const [i, issueToImport] of githubIssuesToImport.entries()) {
    process.stdout.clearLine();
    process.stdout.write(`${chalk.black.bgBlue(' PENDING ')} Transferring Launchpad bugs to GitHub issues. Completed: ${i}/${githubIssuesToImport.length}\r`);

    try {
        const issue = await octokit.request('POST /repos/{owner}/{repo}/import/issues', {
            owner: process.env.REPOSITORY_OWNER,
            repo: process.env.REPOSITORY_NAME,
            ...issueToImport,
            headers: {
                accept: 'application/vnd.github.golden-comet-preview+json',
            },
        });
        pendingIssues.push(issue.data.id);

        if (Number(issue.headers['x-ratelimit-remaining']) < 10) {
            const waitForSec = Math.round((new Date(Number(issue.headers['x-ratelimit-reset']) * 1000) - new Date()) / 1000);
            await sleep(waitForSec * 1000);
            process.stdout.clearLine();
            process.stdout.write(
                `${chalk.black.bgWhite(' SLEEPING ')} Sleeping until the GitHub REST API rate limit resets (${waitForSec} seconds) - Transferring Launchpad bugs to GitHub issues. Completed: ${
                    i + 1
                }/${githubIssuesToImport.length}\r`,
            );
        } else {
            process.stdout.clearLine();
            process.stdout.write(`${chalk.black.bgWhite(' SLEEPING ')} Transferring Launchpad bugs to GitHub issues. Completed: ${i + 1}/${githubIssuesToImport.length}\r`);
            await sleep(1000);
        }
    } catch {
        launchpadBugsThatCouldNotBeImported.push(launchpadBugIds[i]);
        pendingIssues.push(-1);
        process.stdout.clearLine();
        process.stdout.write(`${chalk.black.bgWhite(' SLEEPING ')} Transferring Launchpad bugs to GitHub issues. Completed: ${i + 1}/${githubIssuesToImport.length}\r`);
        await sleep(1000);
    }
}
process.stdout.clearLine();
console.log(
    `${chalk.black.bgGreen(' SUCCESS ')} Transferred ${githubIssuesToImport.length - launchpadBugsThatCouldNotBeImported.length} Launchpad bugs to GitHub issues. ${
        launchpadBugsThatCouldNotBeImported.length
    } bugs could not be transferred (${launchpadBugsThatCouldNotBeImported.join('; ')}).`,
);

// STEP 7: Finalize the transfer of issues by posting an explanatory message on each Launchpad bug that will point to the corresponding imported issue
let issuesWithErrors = [];

async function finalizeIssuesMigration() {
    process.stdout.clearLine();
    process.stdout.write(`${chalk.black.bgWhite(' SLEEPING ')} Waiting for GitHub issue imports to finish. This may take several minutes.\r`);
    await sleep(10 * 1000);

    const issueImportStatus = await octokit.request('GET /repos/{owner}/{repo}/import/issues?{since}', {
        owner: process.env.REPOSITORY_OWNER,
        repo: process.env.REPOSITORY_NAME,
        since: migrationStartTime.toISOString(),
        headers: {
            accept: 'application/vnd.github.golden-comet-preview+json',
        },
    });

    if (Number(issueImportStatus.headers['x-ratelimit-remaining']) < 10) {
        const waitForSec = Math.round((new Date(Number(issueImportStatus.headers['x-ratelimit-reset']) * 1000) - new Date()) / 1000);
        await sleep(waitForSec * 1000);
        process.stdout.clearLine();
        process.stdout.write(
            `${chalk.black.bgWhite(
                ' SLEEPING ',
            )} Sleeping until the GitHub REST API rate limit resets (${waitForSec} seconds) - Waiting for GitHub issue imports to finish. This may take several minutes.\r`,
        );
    }

    // If some issues haven't been imported yet, wait another 10 seconds
    if (issueImportStatus.data.some((issueStatus) => issueStatus.status === 'pending')) await finalizeIssuesMigration();

    for (const [i, issueToFinalize] of issueImportStatus.data.entries()) {
        process.stdout.clearLine();
        process.stdout.write(
            `${chalk.black.bgBlue(' PENDING ')} Posting an informational message on Launchpad bugs that points to the corresponding new GitHub issue. Completed: ${i}/${
                issueImportStatus.data.length
            }\r`,
        );

        if (issueToFinalize.status !== 'imported') {
            issuesWithErrors.push(issueToFinalize.id);
            continue;
        }

        const issueId = Number(issueToFinalize.issue_url.match(/\/issues\/(\d+)/)[1]);
        const launchpadBugId = launchpadBugIds[pendingIssues.indexOf(issueToFinalize.id)];

            /*await fetch(`https://api.launchpad.net/devel/bugs/${launchpadBugId}`, {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
                    Authorization: `OAuth realm="https://api.launchpad.net/",oauth_consumer_key="${encodeURIComponent(LAUNCHPAD_CONSUMER_KEY)}",oauth_token="${accessToken.get(
                        'oauth_token',
                    )}",oauth_signature_method="PLAINTEXT",oauth_signature="%26${accessToken.get('oauth_token_secret')}",oauth_timestamp="${Date.now()}",oauth_nonce="${crypto
                        .randomBytes(16)
                        .toString('hex')}",oauth_version="1.0"`,
                },
                body: new URLSearchParams({
                    'ws.op': 'newMessage',
                    content: `**IMPORTANT**\n\nThis bug has been moved to https://github.com/${process.env.REPOSITORY_OWNER}/${process.env.REPOSITORY_NAME}/issues/${issueId}. Please continue the discussion there.`,
                }),
            });*/

    }

    process.stdout.clearLine();
    console.log(`${chalk.black.bgGreen(' SUCCESS ')} Finalized the transfer of ${issueImportStatus.data.length} Launchpad bugs. ${issuesWithErrors.length} error(s) were encountered.`);
}

await finalizeIssuesMigration();

console.log('\n', chalk.black.bgYellow(' All done, migration complete! '));
process.exit(0);
